/*
  Copyright (C) 2001 Kimmo Pekkola

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
  $Header: //rainbox/cvsroot/Rainmeter/Plugin/MeterHistogram.cpp,v 1.15 2003/02/10 18:12:45 rainy Exp $

  $Log: MeterHistogram.cpp,v $
  Revision 1.15  2003/02/10 18:12:45  rainy
  Now uses GDI+

  Revision 1.14  2002/07/01 15:32:51  rainy
  Removed include to lsapi.h

  Revision 1.13  2002/05/04 08:14:11  rainy
  Histogram now draws the lines one pixel higher.

  Revision 1.12  2002/04/26 18:28:17  rainy
  The meter is not updated if both measures are disabled.

  Revision 1.11  2002/04/26 18:22:02  rainy
  Added possibility to hide the meter.

  Revision 1.10  2002/03/31 09:58:53  rainy
  Added some comments

  Revision 1.9  2001/12/23 10:14:51  rainy
  Hex color values are now also supported.

  Revision 1.8  2001/10/28 10:19:40  rainy
  Fixed a bug with secondary measure not set correctly

  Revision 1.7  2001/10/14 07:32:33  rainy
  In error situations CError is thrown instead just a boolean value.

  Revision 1.6  2001/09/26 16:26:37  rainy
  Small adjustement to the interfaces.
  Implemented BindMeasure()

  Revision 1.5  2001/09/01 12:58:48  rainy
  Removed MaxValues (i.e. they aren't stored anymore).
  Fixed a bug in bitmap histogram placement.

  Revision 1.4  2001/08/25 17:07:28  rainy
  Added support for background images behind the curves.

  Revision 1.3  2001/08/19 09:13:13  rainy
  Invert moved to the measures.

  Revision 1.2  2001/08/12 15:38:54  Rainy
  Adjusted Update()'s interface.
  Added invert for the secondary measure.

  Revision 1.1.1.1  2001/08/11 10:58:19  Rainy
  Added to CVS.

*/

#include "MeterHistogram.h"
#include "Measure.h"
#include "Error.h"

using namespace Gdiplus;

/*
** CMeterHistogram
**
** The constructor
**
*/
CMeterHistogram::CMeterHistogram() : CMeter()
{
	m_SecondaryMeasure = NULL;
	m_PrimaryColor = 0;
	m_SecondaryColor = 0;
	m_BothColor = 0;
	m_MeterPos = 0;
	m_PrimaryBitmap = NULL;
	m_SecondaryBitmap = NULL;
	m_BothBitmap = NULL;
	m_PrimaryValues = NULL;
	m_SecondaryValues = NULL;
	m_Autoscale = NULL;
}

/*
** ~CMeterHistogram
**
** The destructor
**
*/
CMeterHistogram::~CMeterHistogram()
{
	if (m_PrimaryBitmap) delete m_PrimaryBitmap;
	if (m_SecondaryBitmap) delete m_SecondaryBitmap;
	if (m_BothBitmap) delete m_BothBitmap;
	if (m_PrimaryValues) delete [] m_PrimaryValues;
	if (m_SecondaryValues) delete [] m_SecondaryValues;
}

/*
** Initialize
**
** Load the images and calculate the dimensions of the meter from them.
** Or create the brushes if solid color histogram is used.
**
*/
void CMeterHistogram::Initialize(CMeterWindow& meterWindow)
{
	CMeter::Initialize(meterWindow);

	// Load the bitmaps if defined
	if(!m_PrimaryImageName.empty())
	{
		WCHAR* wideSz = ConvertToWide(m_PrimaryImageName.c_str());
		m_PrimaryBitmap = new Bitmap(wideSz);
		delete [] wideSz;

		Status status = m_PrimaryBitmap->GetLastStatus();
		if(Ok != status)
		{
            throw CError(std::string("PrimaryImage not found: ") + m_PrimaryImageName, __LINE__, __FILE__);
		}

		m_W = m_PrimaryBitmap->GetWidth();
		m_H = m_PrimaryBitmap->GetHeight();
	}

	if(!m_SecondaryImageName.empty())
	{
		WCHAR* wideSz = ConvertToWide(m_SecondaryImageName.c_str());
		m_SecondaryBitmap = new Bitmap(wideSz);
		delete [] wideSz;

		Status status = m_SecondaryBitmap->GetLastStatus();
		if(Ok != status)
		{
            throw CError(std::string("SecondaryImage not found: ") + m_SecondaryImageName, __LINE__, __FILE__);
		}
	}

	if(!m_BothImageName.empty())
	{
		WCHAR* wideSz = ConvertToWide(m_BothImageName.c_str());
		m_BothBitmap = new Bitmap(wideSz);
		delete [] wideSz;

		Status status = m_BothBitmap->GetLastStatus();
		if(Ok != status)
		{
            throw CError(std::string("BothImage not found: ") + m_BothImageName, __LINE__, __FILE__);
		}
	}

	// A sanity check
	if (m_SecondaryMeasure && !m_PrimaryImageName.empty() && (m_BothImageName.empty() || m_SecondaryImageName.empty()))
	{
        throw CError(std::string("You need to define SecondaryImage and BothImage also!"), __LINE__, __FILE__);
	}

	// Create buffers for values
	m_PrimaryValues = new UINT[m_W];
	memset(m_PrimaryValues, 0, sizeof(UINT) * m_W);
	if (m_SecondaryMeasure)
	{
		m_SecondaryValues = new UINT[m_W];
		memset(m_SecondaryValues, 0, sizeof(UINT) * m_W);
	}
}

/*
** ReadConfig
**
** Read the meter-specific configs from the ini-file.
**
*/
void CMeterHistogram::ReadConfig(const char* filename, const char* section)
{
	char tmpSz[256];

	// Read common configs
	CMeter::ReadConfig(filename, section);

	// Read configs for Histogram meter
	if(GetPrivateProfileString(section, "PrimaryColor", "0, 255, 0", tmpSz, 255, filename) > 0) 
	{
		m_PrimaryColor = ParseColor(tmpSz);
	}

	if(GetPrivateProfileString(section, "SecondaryColor", "255, 0, 0", tmpSz, 255, filename) > 0) 
	{
		m_SecondaryColor = ParseColor(tmpSz);
	}

	if(GetPrivateProfileString(section, "BothColor", "255, 255, 0", tmpSz, 255, filename) > 0) 
	{
		m_BothColor = ParseColor(tmpSz);
	}

	if(GetPrivateProfileString(section, "SecondaryMeasureName", "", tmpSz, 255, filename) > 0) 
	{
		m_SecondaryMeasureName = tmpSz;
	}

	if(GetPrivateProfileString(section, "PrimaryImage", "", tmpSz, 255, filename) > 0) 
	{
		VarExpansion(tmpSz, tmpSz);		// Expand litestep variables
 		m_PrimaryImageName = tmpSz;
	}

	if(GetPrivateProfileString(section, "SecondaryImage", "", tmpSz, 255, filename) > 0) 
	{
		VarExpansion(tmpSz, tmpSz);		// Expand litestep variables
 		m_SecondaryImageName = tmpSz;
	}

	if(GetPrivateProfileString(section, "BothImage", "", tmpSz, 255, filename) > 0) 
	{
		VarExpansion(tmpSz, tmpSz);		// Expand litestep variables
 		m_BothImageName = tmpSz;
	}

	m_Autoscale = (1 == GetPrivateProfileInt( section, "AutoScale", 0, filename));
}

/*
** Draw
**
** Draws the meter on the double buffer
**
*/
void CMeterHistogram::Draw(CMeterWindow& meterWindow)
{
	if (m_Measure == NULL) return;

	CMeter::Draw(meterWindow);

	// Gather values
	m_PrimaryValues[m_MeterPos] = m_Measure->GetValue();

	if (m_SecondaryMeasure != NULL)
	{
		m_SecondaryValues[m_MeterPos] = m_SecondaryMeasure->GetValue();
	}

	if (!IsHidden()) 
	{
		Graphics graphics(meterWindow.GetDoubleBuffer());
		Pen primaryPen(m_PrimaryColor);
		Pen secondaryPen(m_SecondaryColor);
		Pen bothPen(m_BothColor);
		
		double maxPrimaryValue = m_Measure->GetMaxValue();
		double maxSecondaryValue = 0;
		if (m_SecondaryMeasure) maxSecondaryValue = m_SecondaryMeasure->GetMaxValue();

		if (m_Autoscale)
		{
			int i;
			unsigned int newValue = 0;
			for (i = 0; i != m_W; i++)
			{
				newValue = max(newValue, m_PrimaryValues[i]);
			}

			// Scale the value up to nearest power of 2
			maxPrimaryValue = 2;
			while(maxPrimaryValue < newValue && maxPrimaryValue != 0)
			{
				maxPrimaryValue *= 2;
			}

			if (m_SecondaryMeasure)
			{
				for (i = 0; i != m_W; i++)
				{
					newValue = max(newValue, m_SecondaryValues[i]);
				}

				// Scale the value up to nearest power of 2
				maxSecondaryValue = 2;
				while(maxSecondaryValue < newValue && maxSecondaryValue != 0)
				{
					maxSecondaryValue *= 2;
				}
			}
		}

		for (UINT i = 0; i < m_W; i++)
		{
			double value = ((double)m_PrimaryValues[(i + m_MeterPos + 1) % m_W] / maxPrimaryValue);
			int primaryBarHeight = m_H * value;
			primaryBarHeight = min(m_H, primaryBarHeight);

			if (m_SecondaryMeasure != NULL)
			{
				double value = ((double)m_SecondaryValues[(i + m_MeterPos + 1) % m_W] / maxSecondaryValue);
				int secondaryBarHeight = m_H * value;
				secondaryBarHeight = min(m_H, secondaryBarHeight);

				// Check which measured value is higher
				int bothBarHeight;
				if (secondaryBarHeight > primaryBarHeight)
				{
					bothBarHeight = primaryBarHeight;
				}
				else
				{
					bothBarHeight = secondaryBarHeight;
				}

				// Draw image/color for the both lines
				if (m_PrimaryBitmap)
				{
					Rect r(m_X + i, m_Y + m_H - bothBarHeight, 1, bothBarHeight);
					graphics.DrawImage(m_BothBitmap, r, i, m_H - bothBarHeight, 1, bothBarHeight, UnitPixel);
				}
				else
				{
					graphics.DrawLine(&bothPen, m_X + i, m_Y + m_H, m_X + i, m_Y + m_H - bothBarHeight);
				}

				// Draw the image/color for the rest
				if (secondaryBarHeight > primaryBarHeight)
				{
					if (m_SecondaryBitmap)
					{
						Rect r(m_X + i, m_Y + m_H - secondaryBarHeight, 1, secondaryBarHeight - bothBarHeight);
						graphics.DrawImage(m_SecondaryBitmap, r, i, m_H - secondaryBarHeight, 1, secondaryBarHeight - bothBarHeight, UnitPixel);
					}
					else
					{
						graphics.DrawLine(&secondaryPen, m_X + i, m_Y + m_H - bothBarHeight, m_X + i, m_Y + m_H - secondaryBarHeight);
					}
				}
				else
				{
					if (m_PrimaryBitmap)
					{
						Rect r(m_X + i, m_Y + m_H - primaryBarHeight, 1, primaryBarHeight - bothBarHeight);
						graphics.DrawImage(m_PrimaryBitmap, r, i, m_H - primaryBarHeight, 1, primaryBarHeight - bothBarHeight, UnitPixel);
					}
					else
					{
						graphics.DrawLine(&primaryPen, m_X + i, m_Y + m_H - bothBarHeight, m_X + i, m_Y + m_H - primaryBarHeight);
					}
				}
			}
			else
			{
				if (m_PrimaryBitmap)
				{
					Rect r(m_X + i, m_Y + m_H - primaryBarHeight, 1, primaryBarHeight);
					graphics.DrawImage(m_PrimaryBitmap, r, i, m_H - primaryBarHeight, 1, primaryBarHeight, UnitPixel);
				}
				else
				{
					graphics.DrawLine(&primaryPen, m_X + i, m_Y + m_H, m_X + i, m_Y + m_H - primaryBarHeight);
				}
			}
		}
	}

	m_MeterPos++;
	m_MeterPos %= m_W;
}

/*
** BindMeasure
**
** Overwritten method to handle the secondary measure binding.
**
*/
void CMeterHistogram::BindMeasure(std::list<CMeasure*>& measures)
{
	CMeter::BindMeasure(measures);

	if(!m_SecondaryMeasureName.empty())
	{
		// Go through the list and check it there is a secondary measure for us
		std::list<CMeasure*>::iterator i = measures.begin();
		for( ; i != measures.end(); i++)
		{
			if(_stricmp((*i)->GetName(), m_SecondaryMeasureName.c_str()) == 0)
			{
				m_SecondaryMeasure = (*i);
				return;
			}
		}

        throw CError(std::string("The meter [") + m_Name + "] cannot be bound with [" + m_SecondaryMeasureName + "]!", __LINE__, __FILE__);
	}
}
